"""
GitLab API: https://docs.gitlab.com/ce/api/system_hooks.html
GitLab API: https://docs.gitlab.com/ce/api/groups.html#hooks
GitLab API: https://docs.gitlab.com/ee/api/projects.html#hooks
"""

import re

import pytest
import responses

from gitlab.v4.objects import GroupHook, Hook, ProjectHook

hooks_content = [
    {
        "id": 1,
        "url": "testurl",
        "push_events": True,
        "tag_push_events": True,
    },
    {
        "id": 2,
        "url": "testurl_second",
        "push_events": False,
        "tag_push_events": False,
    },
]

hook_content = hooks_content[0]


@pytest.fixture
def resp_hooks_list():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=re.compile(r"http://localhost/api/v4/((groups|projects)/1/|)hooks"),
            json=hooks_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_hook_get():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=re.compile(r"http://localhost/api/v4/((groups|projects)/1/|)hooks/1"),
            json=hook_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_hook_create():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url=re.compile(r"http://localhost/api/v4/((groups|projects)/1/|)hooks"),
            json=hook_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_hook_update():
    with responses.RequestsMock() as rsps:
        pattern = re.compile(r"http://localhost/api/v4/((groups|projects)/1/|)hooks/1")
        rsps.add(
            method=responses.GET,
            url=pattern,
            json=hook_content,
            content_type="application/json",
            status=200,
        )
        rsps.add(
            method=responses.PUT,
            url=pattern,
            json=hook_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_hook_delete():
    with responses.RequestsMock() as rsps:
        pattern = re.compile(r"http://localhost/api/v4/((groups|projects)/1/|)hooks/1")
        rsps.add(
            method=responses.GET,
            url=pattern,
            json=hook_content,
            content_type="application/json",
            status=200,
        )
        rsps.add(
            method=responses.DELETE,
            url=pattern,
            status=204,
        )
        yield rsps


def test_list_system_hooks(gl, resp_hooks_list):
    hooks = gl.hooks.list()
    assert hooks[0].id == 1
    assert hooks[0].url == "testurl"
    assert hooks[1].id == 2
    assert hooks[1].url == "testurl_second"


def test_get_system_hook(gl, resp_hook_get):
    data = gl.hooks.get(1)
    assert isinstance(data, Hook)
    assert data.url == "testurl"
    assert data.id == 1


def test_create_system_hook(gl, resp_hook_create):
    hook = gl.hooks.create(hook_content)
    assert hook.url == "testurl"
    assert hook.push_events is True
    assert hook.tag_push_events is True


# there is no update method for system hooks


def test_delete_system_hook(gl, resp_hook_delete):
    hook = gl.hooks.get(1)
    hook.delete()
    gl.hooks.delete(1)


def test_list_group_hooks(group, resp_hooks_list):
    hooks = group.hooks.list()
    assert hooks[0].id == 1
    assert hooks[0].url == "testurl"
    assert hooks[1].id == 2
    assert hooks[1].url == "testurl_second"


def test_get_group_hook(group, resp_hook_get):
    data = group.hooks.get(1)
    assert isinstance(data, GroupHook)
    assert data.url == "testurl"
    assert data.id == 1


def test_create_group_hook(group, resp_hook_create):
    hook = group.hooks.create(hook_content)
    assert hook.url == "testurl"
    assert hook.push_events is True
    assert hook.tag_push_events is True


def test_update_group_hook(group, resp_hook_update):
    hook = group.hooks.get(1)
    assert hook.id == 1
    hook.url = "testurl_more"
    hook.save()


def test_delete_group_hook(group, resp_hook_delete):
    hook = group.hooks.get(1)
    hook.delete()
    group.hooks.delete(1)


def test_list_project_hooks(project, resp_hooks_list):
    hooks = project.hooks.list()
    assert hooks[0].id == 1
    assert hooks[0].url == "testurl"
    assert hooks[1].id == 2
    assert hooks[1].url == "testurl_second"


def test_get_project_hook(project, resp_hook_get):
    data = project.hooks.get(1)
    assert isinstance(data, ProjectHook)
    assert data.url == "testurl"
    assert data.id == 1


def test_create_project_hook(project, resp_hook_create):
    hook = project.hooks.create(hook_content)
    assert hook.url == "testurl"
    assert hook.push_events is True
    assert hook.tag_push_events is True


def test_update_project_hook(project, resp_hook_update):
    hook = project.hooks.get(1)
    assert hook.id == 1
    hook.url = "testurl_more"
    hook.save()


def test_delete_project_hook(project, resp_hook_delete):
    hook = project.hooks.get(1)
    hook.delete()
    project.hooks.delete(1)



"""
Some test cases are run in-process to intercept requests to gitlab.com
and example servers.
"""

import copy
import json

import pytest
import responses
import yaml

from gitlab import __version__, config
from gitlab.const import DEFAULT_URL

PRIVATE_TOKEN = "glpat-abc123"
CI_JOB_TOKEN = "ci-job-token"
CI_SERVER_URL = "https://gitlab.example.com"


def test_main_entrypoint(script_runner, gitlab_config):
    ret = script_runner.run(["python", "-m", "gitlab", "--config-file", gitlab_config])
    assert ret.returncode == 2


def test_version(script_runner):
    ret = script_runner.run(["gitlab", "--version"])
    assert ret.stdout.strip() == __version__


def test_config_error_with_help_prints_help(script_runner):
    ret = script_runner.run(["gitlab", "-c", "invalid-file", "--help"])
    assert ret.stdout.startswith("usage:")
    assert ret.returncode == 0


def test_resource_help_prints_actions_vertically(script_runner):
    ret = script_runner.run(["gitlab", "project", "--help"])
    assert "    list                List the GitLab resources\n" in ret.stdout
    assert "    get                 Get a GitLab resource\n" in ret.stdout
    assert ret.returncode == 0


def test_resource_help_prints_actions_vertically_only_one_action(script_runner):
    ret = script_runner.run(["gitlab", "event", "--help"])
    assert "  {list}      Action to execute on the GitLab resource.\n"
    assert "    list      List the GitLab resources\n" in ret.stdout
    assert ret.returncode == 0


@pytest.mark.script_launch_mode("inprocess")
@responses.activate
def test_defaults_to_gitlab_com(script_runner, resp_get_project, monkeypatch):
    responses.add(**resp_get_project)
    monkeypatch.setattr(config, "_DEFAULT_FILES", [])
    ret = script_runner.run(["gitlab", "project", "get", "--id", "1"])
    assert ret.success
    assert "id: 1" in ret.stdout


@pytest.mark.script_launch_mode("inprocess")
@responses.activate
def test_uses_ci_server_url(monkeypatch, script_runner, resp_get_project):
    monkeypatch.setenv("CI_SERVER_URL", CI_SERVER_URL)
    monkeypatch.setattr(config, "_DEFAULT_FILES", [])
    resp_get_project_in_ci = copy.deepcopy(resp_get_project)
    resp_get_project_in_ci.update(url=f"{CI_SERVER_URL}/api/v4/projects/1")

    responses.add(**resp_get_project_in_ci)
    ret = script_runner.run(["gitlab", "project", "get", "--id", "1"])
    assert ret.success


@pytest.mark.script_launch_mode("inprocess")
@responses.activate
def test_uses_ci_job_token(monkeypatch, script_runner, resp_get_project):
    monkeypatch.setenv("CI_JOB_TOKEN", CI_JOB_TOKEN)
    monkeypatch.setattr(config, "_DEFAULT_FILES", [])
    resp_get_project_in_ci = copy.deepcopy(resp_get_project)
    resp_get_project_in_ci.update(
        match=[responses.matchers.header_matcher({"JOB-TOKEN": CI_JOB_TOKEN})],
    )

    responses.add(**resp_get_project_in_ci)
    ret = script_runner.run(["gitlab", "project", "get", "--id", "1"])
    assert ret.success


@pytest.mark.script_launch_mode("inprocess")
@responses.activate
def test_does_not_auth_on_skip_login(
    monkeypatch, script_runner, resp_get_project, resp_current_user
):
    monkeypatch.setenv("GITLAB_PRIVATE_TOKEN", PRIVATE_TOKEN)
    monkeypatch.setattr(config, "_DEFAULT_FILES", [])

    resp_user = responses.add(**resp_current_user)
    resp_project = responses.add(**resp_get_project)
    ret = script_runner.run(["gitlab", "--skip-login", "project", "get", "--id", "1"])
    assert ret.success
    assert resp_user.call_count == 0
    assert resp_project.call_count == 1


@pytest.mark.script_launch_mode("inprocess")
@responses.activate
def test_private_token_overrides_job_token(
    monkeypatch, script_runner, resp_get_project
):
    monkeypatch.setenv("GITLAB_PRIVATE_TOKEN", PRIVATE_TOKEN)
    monkeypatch.setenv("CI_JOB_TOKEN", CI_JOB_TOKEN)

    resp_get_project_with_token = copy.deepcopy(resp_get_project)
    resp_get_project_with_token.update(
        match=[responses.matchers.header_matcher({"PRIVATE-TOKEN": PRIVATE_TOKEN})],
    )

    # CLI first calls .auth() when private token is present
    resp_auth_with_token = copy.deepcopy(resp_get_project_with_token)
    resp_auth_with_token.update(url=f"{DEFAULT_URL}/api/v4/user")
    resp_auth_with_token["json"].update(username="user", web_url=f"{DEFAULT_URL}/user")

    responses.add(**resp_get_project_with_token)
    responses.add(**resp_auth_with_token)
    ret = script_runner.run(["gitlab", "project", "get", "--id", "1"])
    assert ret.success


def test_env_config_missing_file_raises(script_runner, monkeypatch):
    monkeypatch.setenv("PYTHON_GITLAB_CFG", "non-existent")
    ret = script_runner.run(["gitlab", "project", "list"])
    assert not ret.success
    assert ret.stderr.startswith("Cannot read config from PYTHON_GITLAB_CFG")


def test_arg_config_missing_file_raises(script_runner):
    ret = script_runner.run(
        ["gitlab", "--config-file", "non-existent", "project", "list"]
    )
    assert not ret.success
    assert ret.stderr.startswith("Cannot read config from file")


def test_invalid_config(script_runner):
    ret = script_runner.run(["gitlab", "--gitlab", "invalid"])
    assert not ret.success
    assert not ret.stdout


def test_invalid_config_prints_help(script_runner):
    ret = script_runner.run(["gitlab", "--gitlab", "invalid", "--help"])
    assert ret.success
    assert ret.stdout


def test_invalid_api_version(script_runner, monkeypatch, fixture_dir):
    monkeypatch.setenv("PYTHON_GITLAB_CFG", str(fixture_dir / "invalid_version.cfg"))
    ret = script_runner.run(["gitlab", "--gitlab", "test", "project", "list"])
    assert not ret.success
    assert ret.stderr.startswith("Unsupported API version:")


def test_invalid_auth_config(script_runner, monkeypatch, fixture_dir):
    monkeypatch.setenv("PYTHON_GITLAB_CFG", str(fixture_dir / "invalid_auth.cfg"))
    ret = script_runner.run(["gitlab", "--gitlab", "test", "project", "list"])
    assert not ret.success
    assert "401" in ret.stderr


format_matrix = [
    ("json", json.loads),
    ("yaml", yaml.safe_load),
]


@pytest.mark.parametrize("format,loader", format_matrix)
def test_cli_display(gitlab_cli, project, format, loader):
    cmd = ["-o", format, "project", "get", "--id", project.id]

    ret = gitlab_cli(cmd)
    assert ret.success

    content = loader(ret.stdout.strip())
    assert content["id"] == project.id


@pytest.mark.parametrize("format,loader", format_matrix)
def test_cli_fields_in_list(gitlab_cli, project_file, format, loader):
    cmd = [
        "-o",
        format,
        "--fields",
        "default_branch",
        "project",
        "list",
    ]

    ret = gitlab_cli(cmd)
    assert ret.success

    content = loader(ret.stdout.strip())
    assert ["default_branch" in item for item in content]


def test_cli_display_without_fields_warns(gitlab_cli, project):
    cmd = ["project-ci-lint", "get", "--project-id", project.id]

    ret = gitlab_cli(cmd)
    assert ret.success

    assert "No default fields to show" in ret.stdout
    assert "merged_yaml" not in ret.stdout


def test_cli_does_not_print_token(gitlab_cli, gitlab_token):
    ret = gitlab_cli(["--debug", "current-user", "get"])
    assert ret.success

    assert gitlab_token not in ret.stdout
    assert gitlab_token not in ret.stderr
    assert "[MASKED]" in ret.stderr



from typing import Any, cast, Union

from gitlab.base import RESTManager, RESTObject
from gitlab.mixins import CRUDMixin, ObjectDeleteMixin, SaveMixin, UploadMixin
from gitlab.types import RequiredOptional

__all__ = [
    "ProjectWiki",
    "ProjectWikiManager",
    "GroupWiki",
    "GroupWikiManager",
]


class ProjectWiki(SaveMixin, ObjectDeleteMixin, UploadMixin, RESTObject):
    _id_attr = "slug"
    _repr_attr = "slug"
    _upload_path = "/projects/{project_id}/wikis/attachments"


class ProjectWikiManager(CRUDMixin, RESTManager):
    _path = "/projects/{project_id}/wikis"
    _obj_cls = ProjectWiki
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(
        required=("title", "content"), optional=("format",)
    )
    _update_attrs = RequiredOptional(optional=("title", "content", "format"))
    _list_filters = ("with_content",)

    def get(
        self, id: Union[str, int], lazy: bool = False, **kwargs: Any
    ) -> ProjectWiki:
        return cast(ProjectWiki, super().get(id=id, lazy=lazy, **kwargs))


class GroupWiki(SaveMixin, ObjectDeleteMixin, UploadMixin, RESTObject):
    _id_attr = "slug"
    _repr_attr = "slug"
    _upload_path = "/groups/{group_id}/wikis/attachments"


class GroupWikiManager(CRUDMixin, RESTManager):
    _path = "/groups/{group_id}/wikis"
    _obj_cls = GroupWiki
    _from_parent_attrs = {"group_id": "id"}
    _create_attrs = RequiredOptional(
        required=("title", "content"), optional=("format",)
    )
    _update_attrs = RequiredOptional(optional=("title", "content", "format"))
    _list_filters = ("with_content",)

    def get(self, id: Union[str, int], lazy: bool = False, **kwargs: Any) -> GroupWiki:
        return cast(GroupWiki, super().get(id=id, lazy=lazy, **kwargs))



"""
Ensure type-hints are setup correctly and detect if missing functions.

Original notes by John L. Villalovos

"""

import dataclasses
import functools
import inspect
from typing import Optional, Type

import pytest

import gitlab.mixins
import gitlab.v4.objects


@functools.total_ordering
@dataclasses.dataclass(frozen=True)
class ClassInfo:
    name: str
    type: Type  # type: ignore[type-arg]

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, ClassInfo):
            return NotImplemented
        return (self.type.__module__, self.name) < (other.type.__module__, other.name)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ClassInfo):
            return NotImplemented
        return (self.type.__module__, self.name) == (other.type.__module__, other.name)


def pytest_generate_tests(metafunc: pytest.Metafunc) -> None:
    """Find all of the classes in gitlab.v4.objects and pass them to our test
    function"""

    class_info_set = set()
    for _, module_value in inspect.getmembers(gitlab.v4.objects):
        if not inspect.ismodule(module_value):
            # We only care about the modules
            continue
        # Iterate through all the classes in our module
        for class_name, class_value in inspect.getmembers(module_value):
            if not inspect.isclass(class_value):
                continue

            module_name = class_value.__module__
            # Ignore imported classes from gitlab.base
            if module_name == "gitlab.base":
                continue

            if not class_name.endswith("Manager"):
                continue

            class_info_set.add(ClassInfo(name=class_name, type=class_value))

    metafunc.parametrize("class_info", sorted(class_info_set))


GET_ID_METHOD_TEMPLATE = """
def get(
    self, id: Union[str, int], lazy: bool = False, **kwargs: Any
) -> {obj_cls.__name__}:
    return cast({obj_cls.__name__}, super().get(id=id, lazy=lazy, **kwargs))

You may also need to add the following imports:
from typing import Any, cast, Union"
"""

GET_WITHOUT_ID_METHOD_TEMPLATE = """
def get(self, **kwargs: Any) -> {obj_cls.__name__}:
    return cast({obj_cls.__name__}, super().get(**kwargs))

You may also need to add the following imports:
from typing import Any, cast"
"""


class TestTypeHints:
    def test_check_get_function_type_hints(self, class_info: ClassInfo) -> None:
        """Ensure classes derived from GetMixin have defined a 'get()' method with
        correct type-hints.
        """
        self.get_check_helper(
            base_type=gitlab.mixins.GetMixin,
            class_info=class_info,
            method_template=GET_ID_METHOD_TEMPLATE,
            optional_return=False,
        )

    def test_check_get_without_id_function_type_hints(
        self, class_info: ClassInfo
    ) -> None:
        """Ensure classes derived from GetMixin have defined a 'get()' method with
        correct type-hints.
        """
        self.get_check_helper(
            base_type=gitlab.mixins.GetWithoutIdMixin,
            class_info=class_info,
            method_template=GET_WITHOUT_ID_METHOD_TEMPLATE,
            optional_return=False,
        )

    def get_check_helper(
        self,
        *,
        base_type: Type,  # type: ignore[type-arg]
        class_info: ClassInfo,
        method_template: str,
        optional_return: bool,
    ) -> None:
        if not class_info.name.endswith("Manager"):
            return
        mro = class_info.type.mro()
        # The class needs to be derived from GetMixin or we ignore it
        if base_type not in mro:
            return

        obj_cls = class_info.type._obj_cls
        signature = inspect.signature(class_info.type.get)
        filename = inspect.getfile(class_info.type)

        fail_message = (
            f"class definition for {class_info.name!r} in file {filename!r} "
            f"must have defined a 'get' method with a return annotation of "
            f"{obj_cls} but found {signature.return_annotation}\n"
            f"Recommend adding the following method:\n"
        )
        fail_message += method_template.format(obj_cls=obj_cls)
        check_type = obj_cls
        if optional_return:
            check_type = Optional[obj_cls]
        assert check_type == signature.return_annotation, fail_message



import inspect
import os

import jinja2
import sphinx
import sphinx.ext.napoleon as napoleon
from sphinx.ext.napoleon.docstring import GoogleDocstring


def classref(value, short=True):
    return value

    if not inspect.isclass(value):
        return f":class:{value}"
    tilde = "~" if short else ""
    return f":class:`{tilde}gitlab.objects.{value.__name__}`"


def setup(app):
    app.connect("autodoc-process-docstring", _process_docstring)
    app.connect("autodoc-skip-member", napoleon._skip_member)

    conf = napoleon.Config._config_values

    for name, (default, rebuild) in conf.items():
        app.add_config_value(name, default, rebuild)
    return {"version": sphinx.__display_version__, "parallel_read_safe": True}


def _process_docstring(app, what, name, obj, options, lines):
    result_lines = lines
    docstring = GitlabDocstring(result_lines, app.config, app, what, name, obj, options)
    result_lines = docstring.lines()
    lines[:] = result_lines[:]


class GitlabDocstring(GoogleDocstring):
    def _build_doc(self, tmpl, **kwargs):
        env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(os.path.dirname(__file__)), trim_blocks=False
        )
        env.filters["classref"] = classref
        template = env.get_template(tmpl)
        output = template.render(**kwargs)

        return output.split("\n")

    def __init__(
        self, docstring, config=None, app=None, what="", name="", obj=None, options=None
    ):
        super().__init__(docstring, config, app, what, name, obj, options)

        if name.startswith("gitlab.v4.objects") and name.endswith("Manager"):
            self._parsed_lines.extend(self._build_doc("manager_tmpl.j2", cls=self._obj))



# https://docs.gitlab.com/ee/ci/jobs/ci_job_token.html#allow-any-project-to-access-your-project
def test_enable_limit_access_to_this_project(gl, project):
    scope = project.job_token_scope.get()

    scope.enabled = True
    scope.save()

    scope.refresh()

    assert scope.inbound_enabled


def test_disable_limit_access_to_this_project(gl, project):
    scope = project.job_token_scope.get()

    scope.enabled = False
    scope.save()

    scope.refresh()

    assert not scope.inbound_enabled


def test_add_project_to_job_token_scope_allowlist(gl, project):
    project_to_add = gl.projects.create({"name": "Ci_Cd_token_add_proj"})

    scope = project.job_token_scope.get()
    resp = scope.allowlist.create({"target_project_id": project_to_add.id})

    assert resp.source_project_id == project.id
    assert resp.target_project_id == project_to_add.id

    project_to_add.delete()


def test_projects_job_token_scope_allowlist_contains_added_project_name(gl, project):
    scope = project.job_token_scope.get()
    project_name = "Ci_Cd_token_named_proj"
    project_to_add = gl.projects.create({"name": project_name})
    scope.allowlist.create({"target_project_id": project_to_add.id})

    scope.refresh()
    assert any(allowed.name == project_name for allowed in scope.allowlist.list())

    project_to_add.delete()


def test_remove_project_by_id_from_projects_job_token_scope_allowlist(gl, project):
    scope = project.job_token_scope.get()

    project_to_add = gl.projects.create({"name": "Ci_Cd_token_remove_proj"})

    scope.allowlist.create({"target_project_id": project_to_add.id})

    scope.refresh()

    scope.allowlist.delete(project_to_add.id)

    scope.refresh()
    assert not any(
        allowed.id == project_to_add.id for allowed in scope.allowlist.list()
    )

    project_to_add.delete()


def test_add_group_to_job_token_scope_allowlist(gl, project):
    group_to_add = gl.groups.create(
        {"name": "add_group", "path": "allowlisted-add-test"}
    )

    scope = project.job_token_scope.get()
    resp = scope.groups_allowlist.create({"target_group_id": group_to_add.id})

    assert resp.source_project_id == project.id
    assert resp.target_group_id == group_to_add.id

    group_to_add.delete()


def test_projects_job_token_scope_groups_allowlist_contains_added_group_name(
    gl, project
):
    scope = project.job_token_scope.get()
    group_name = "list_group"
    group_to_add = gl.groups.create(
        {"name": group_name, "path": "allowlisted-add-and-list-test"}
    )

    scope.groups_allowlist.create({"target_group_id": group_to_add.id})

    scope.refresh()
    assert any(allowed.name == group_name for allowed in scope.groups_allowlist.list())

    group_to_add.delete()


def test_remove_group_by_id_from_projects_job_token_scope_groups_allowlist(gl, project):
    scope = project.job_token_scope.get()

    group_to_add = gl.groups.create(
        {"name": "delete_group", "path": "allowlisted-delete-test"}
    )

    scope.groups_allowlist.create({"target_group_id": group_to_add.id})

    scope.refresh()

    scope.groups_allowlist.delete(group_to_add.id)

    scope.refresh()
    assert not any(
        allowed.name == group_to_add.name for allowed in scope.groups_allowlist.list()
    )

    group_to_add.delete()



import json
import time


def test_project_create_file(gitlab_cli, project):
    file_path = "README"
    branch = "main"
    content = "CONTENT"
    commit_message = "Initial commit"

    cmd = [
        "project-file",
        "create",
        "--project-id",
        project.id,
        "--file-path",
        file_path,
        "--branch",
        branch,
        "--content",
        content,
        "--commit-message",
        commit_message,
    ]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_list_all_commits(gitlab_cli, project):
    data = {
        "branch": "new-branch",
        "start_branch": "main",
        "commit_message": "chore: test commit on new branch",
        "actions": [
            {
                "action": "create",
                "file_path": "test-cli-repo.md",
                "content": "new content",
            }
        ],
    }
    commit = project.commits.create(data)

    cmd = ["project-commit", "list", "--project-id", project.id, "--get-all"]
    ret = gitlab_cli(cmd)
    assert commit.id not in ret.stdout

    # Listing commits on other branches requires `all` parameter passed to the API
    cmd = [
        "project-commit",
        "list",
        "--project-id",
        project.id,
        "--get-all",
        "--all",
        "true",
    ]
    ret_all = gitlab_cli(cmd)
    assert commit.id in ret_all.stdout
    assert len(ret_all.stdout) > len(ret.stdout)


def test_list_merge_request_commits(gitlab_cli, merge_request, project):
    cmd = [
        "project-merge-request",
        "commits",
        "--project-id",
        project.id,
        "--iid",
        merge_request.iid,
    ]

    ret = gitlab_cli(cmd)
    assert ret.success
    assert ret.stdout


def test_commit_merge_requests(gitlab_cli, project, merge_request):
    """This tests the `project-commit merge-requests` command and also tests
    that we can print the result using the `json` formatter"""

    # Pause to let GL catch up (happens on hosted too, sometimes takes a while for server to be ready to merge)
    time.sleep(30)

    merge_result = merge_request.merge(should_remove_source_branch=True)
    # Pause to let GL catch up (happens on hosted too, sometimes takes a while for server to be ready to merge)
    time.sleep(5)

    # Wait until it is merged
    mr = None
    mr_iid = merge_request.iid
    for _ in range(60):
        mr = project.mergerequests.get(mr_iid)
        if mr.merged_at is not None:
            break
        time.sleep(0.5)

    assert mr is not None
    assert mr.merged_at is not None
    # Pause to let GL catch up (happens on hosted too, sometimes takes a while for server to be ready to merge)
    time.sleep(5)

    commit_sha = merge_result["sha"]
    cmd = [
        "-o",
        "json",
        "project-commit",
        "merge-requests",
        "--project-id",
        project.id,
        "--id",
        commit_sha,
    ]
    ret = gitlab_cli(cmd)
    assert ret.success

    json_list = json.loads(ret.stdout)
    assert isinstance(json_list, list)
    assert len(json_list) == 1
    mr_dict = json_list[0]
    assert mr_dict["id"] == mr.id
    assert mr_dict["iid"] == mr.iid


def test_revert_commit(gitlab_cli, project):
    commit = project.commits.list()[0]

    cmd = [
        "project-commit",
        "revert",
        "--project-id",
        project.id,
        "--id",
        commit.id,
        "--branch",
        "main",
    ]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_get_commit_signature_not_found(gitlab_cli, project):
    commit = project.commits.list()[0]

    cmd = ["project-commit", "signature", "--project-id", project.id, "--id", commit.id]
    ret = gitlab_cli(cmd)

    assert not ret.success
    assert "404 Signature Not Found" in ret.stderr



from typing import Any, cast, Union

from gitlab.base import RESTManager, RESTObject
from gitlab.mixins import CRUDMixin, ObjectDeleteMixin, SaveMixin
from gitlab.types import ArrayAttribute, RequiredOptional

__all__ = [
    "BroadcastMessage",
    "BroadcastMessageManager",
]


class BroadcastMessage(SaveMixin, ObjectDeleteMixin, RESTObject):
    pass


class BroadcastMessageManager(CRUDMixin, RESTManager):
    _path = "/broadcast_messages"
    _obj_cls = BroadcastMessage

    _create_attrs = RequiredOptional(
        required=("message",),
        optional=("starts_at", "ends_at", "color", "font", "target_access_levels"),
    )
    _update_attrs = RequiredOptional(
        optional=(
            "message",
            "starts_at",
            "ends_at",
            "color",
            "font",
            "target_access_levels",
        )
    )
    _types = {"target_access_levels": ArrayAttribute}

    def get(
        self, id: Union[str, int], lazy: bool = False, **kwargs: Any
    ) -> BroadcastMessage:
        return cast(BroadcastMessage, super().get(id=id, lazy=lazy, **kwargs))



from typing import Any, cast, Union

from gitlab.base import RESTManager, RESTObject
from gitlab.mixins import NoUpdateMixin, ObjectDeleteMixin, SaveMixin
from gitlab.types import RequiredOptional

__all__ = [
    "ProjectClusterAgent",
    "ProjectClusterAgentManager",
]


class ProjectClusterAgent(SaveMixin, ObjectDeleteMixin, RESTObject):
    _repr_attr = "name"


class ProjectClusterAgentManager(NoUpdateMixin, RESTManager):
    _path = "/projects/{project_id}/cluster_agents"
    _obj_cls = ProjectClusterAgent
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(required=("name",))

    def get(
        self, id: Union[str, int], lazy: bool = False, **kwargs: Any
    ) -> ProjectClusterAgent:
        return cast(ProjectClusterAgent, super().get(id=id, lazy=lazy, **kwargs))



import pytest
from requests_toolbelt.multipart.encoder import MultipartEncoder  # type: ignore

from gitlab._backends import requests_backend


class TestSendData:
    def test_senddata_json(self) -> None:
        result = requests_backend.SendData(
            json={"a": 1}, content_type="application/json"
        )
        assert result.data is None

    def test_senddata_data(self) -> None:
        result = requests_backend.SendData(
            data={"b": 2}, content_type="application/octet-stream"
        )
        assert result.json is None

    def test_senddata_json_and_data(self) -> None:
        with pytest.raises(ValueError, match=r"json={'a': 1}  data={'b': 2}"):
            requests_backend.SendData(
                json={"a": 1}, data={"b": 2}, content_type="application/json"
            )


class TestRequestsBackend:
    @pytest.mark.parametrize(
        "test_data,expected",
        [
            (False, "0"),
            (True, "1"),
            ("12", "12"),
            (12, "12"),
            (12.0, "12.0"),
            (complex(-2, 7), "(-2+7j)"),
        ],
    )
    def test_prepare_send_data_non_strings(self, test_data, expected) -> None:
        assert isinstance(expected, str)
        files = {"file": ("file.tar.gz", "12345", "application/octet-stream")}
        post_data = {"test_data": test_data}

        result = requests_backend.RequestsBackend.prepare_send_data(
            files=files, post_data=post_data, raw=False
        )
        assert result.json is None
        assert result.content_type.startswith("multipart/form-data")
        assert isinstance(result.data, MultipartEncoder)
        assert isinstance(result.data.fields["test_data"], str)
        assert result.data.fields["test_data"] == expected



"""
GitLab API:
https://docs.gitlab.com/ee/api/branches.html
"""


def test_branch_name_with_period(project):
    # Make sure we can create and get a branch name containing a period '.'
    branch_name = "my.branch.name"
    branch = project.branches.create({"branch": branch_name, "ref": "main"})
    assert branch.name == branch_name

    # Ensure we can get the branch
    fetched_branch = project.branches.get(branch_name)
    assert branch.name == fetched_branch.name

    branch.delete()



import io
import sys
from pathlib import Path
from textwrap import dedent
from unittest import mock

import pytest

import gitlab
from gitlab import config, const

custom_user_agent = "my-package/1.0.0"

valid_config = """[global]
default = one
ssl_verify = true
timeout = 2

[one]
url = http://one.url
private_token = ABCDEF

[two]
url = https://two.url
private_token = GHIJKL
ssl_verify = false
timeout = 10

[three]
url = https://three.url
private_token = MNOPQR
ssl_verify = /path/to/CA/bundle.crt
per_page = 50

[four]
url = https://four.url
oauth_token = STUV
"""

ssl_verify_str_config = """[global]
default = one
ssl_verify = /etc/ssl/certs/ca-certificates.crt

[one]
url = http://one.url
private_token = ABCDEF
"""

custom_user_agent_config = f"""[global]
default = one
user_agent = {custom_user_agent}

[one]
url = http://one.url
private_token = ABCDEF
"""

no_default_config = """[global]
[there]
url = http://there.url
private_token = ABCDEF
"""

invalid_data_config = """[global]
[one]
url = http://one.url

[two]
private_token = ABCDEF

[three]
meh = hem

[four]
url = http://four.url
private_token = ABCDEF
per_page = 200

[invalid-api-version]
url = http://invalid-api-version.url
private_token = ABCDEF
api_version = 1
"""


@pytest.fixture(autouse=True)
def default_files(monkeypatch):
    """Overrides mocked default files from conftest.py as we have our own mocks here."""
    monkeypatch.setattr(gitlab.config, "_DEFAULT_FILES", config._DEFAULT_FILES)


def global_retry_transient_errors(value: bool) -> str:
    return f"""[global]
default = one
retry_transient_errors={value}
[one]
url = http://one.url
private_token = ABCDEF"""


def global_and_gitlab_retry_transient_errors(
    global_value: bool, gitlab_value: bool
) -> str:
    return f"""[global]
    default = one
    retry_transient_errors={global_value}
    [one]
    url = http://one.url
    private_token = ABCDEF
    retry_transient_errors={gitlab_value}"""


def _mock_nonexistent_file(*args, **kwargs):
    raise OSError


def _mock_existent_file(path, *args, **kwargs):
    return path


def test_env_config_missing_file_raises(monkeypatch):
    monkeypatch.setenv("PYTHON_GITLAB_CFG", "/some/path")
    with pytest.raises(config.GitlabConfigMissingError):
        config._get_config_files()


def test_env_config_not_defined_does_not_raise(monkeypatch):
    with monkeypatch.context() as m:
        m.setattr(config, "_DEFAULT_FILES", [])
        assert config._get_config_files() == []


def test_default_config(monkeypatch):
    with monkeypatch.context() as m:
        m.setattr(Path, "resolve", _mock_nonexistent_file)
        cp = config.GitlabConfigParser()

    assert cp.gitlab_id is None
    assert cp.http_username is None
    assert cp.http_password is None
    assert cp.job_token is None
    assert cp.oauth_token is None
    assert cp.private_token is None
    assert cp.api_version == "4"
    assert cp.order_by is None
    assert cp.pagination is None
    assert cp.per_page is None
    assert cp.retry_transient_errors is False
    assert cp.ssl_verify is True
    assert cp.timeout == 60
    assert cp.url is None
    assert cp.user_agent == const.USER_AGENT


@mock.patch("builtins.open")
def test_invalid_id(m_open, monkeypatch):
    fd = io.StringIO(no_default_config)
    fd.close = mock.Mock(return_value=None)
    m_open.return_value = fd
    with monkeypatch.context() as m:
        m.setattr(Path, "resolve", _mock_existent_file)
        config.GitlabConfigParser("there")
        with pytest.raises(config.GitlabIDError):
            config.GitlabConfigParser()
        fd = io.StringIO(valid_config)
        fd.close = mock.Mock(return_value=None)
        m_open.return_value = fd
        with pytest.raises(config.GitlabDataError):
            config.GitlabConfigParser(gitlab_id="not_there")


@mock.patch("builtins.open")
def test_invalid_data(m_open, monkeypatch):
    fd = io.StringIO(invalid_data_config)
    fd.close = mock.Mock(return_value=None, side_effect=lambda: fd.seek(0))
    m_open.return_value = fd

    with monkeypatch.context() as m:
        m.setattr(Path, "resolve", _mock_existent_file)
        config.GitlabConfigParser("one")
        config.GitlabConfigParser("one")
        with pytest.raises(config.GitlabDataError):
            config.GitlabConfigParser(gitlab_id="two")
        with pytest.raises(config.GitlabDataError):
            config.GitlabConfigParser(gitlab_id="three")

        with pytest.raises(config.GitlabDataError) as e:
            config.GitlabConfigParser("four")
        assert str(e.value) == "Unsupported per_page number: 200"

        with pytest.raises(config.GitlabDataError) as e:
            config.GitlabConfigParser("invalid-api-version")
        assert str(e.value) == "Unsupported API version: 1"


@mock.patch("builtins.open")
def test_valid_data(m_open, monkeypatch):
    fd = io.StringIO(valid_config)
    fd.close = mock.Mock(return_value=None)
    m_open.return_value = fd

    with monkeypatch.context() as m:
        m.setattr(Path, "resolve", _mock_existent_file)
        cp = config.GitlabConfigParser()
    assert "one" == cp.gitlab_id
    assert "http://one.url" == cp.url
    assert "ABCDEF" == cp.private_token
    assert cp.oauth_token is None
    assert 2 == cp.timeout
    assert cp.ssl_verify is True
    assert cp.per_page is None

    fd = io.StringIO(valid_config)
    fd.close = mock.Mock(return_value=None)
    m_open.return_value = fd
    with monkeypatch.context() as m:
        m.setattr(Path, "resolve", _mock_existent_file)
        cp = config.GitlabConfigParser(gitlab_id="two")
    assert "two" == cp.gitlab_id
    assert "https://two.url" == cp.url
    assert "GHIJKL" == cp.private_token
    assert cp.oauth_token is None
    assert 10 == cp.timeout
    assert cp.ssl_verify is False

    fd = io.StringIO(valid_config)
    fd.close = mock.Mock(return_value=None)
    m_open.return_value = fd
    with monkeypatch.context() as m:
        m.setattr(Path, "resolve", _mock_existent_file)
        cp = config.GitlabConfigParser(gitlab_id="three")
    assert "three" == cp.gitlab_id
    assert "https://three.url" == cp.url
    assert "MNOPQR" == cp.private_token
    assert cp.oauth_token is None
    assert 2 == cp.timeout
    assert "/path/to/CA/bundle.crt" == cp.ssl_verify
    assert 50 == cp.per_page

    fd = io.StringIO(valid_config)
    fd.close = mock.Mock(return_value=None)
    m_open.return_value = fd
    with monkeypatch.context() as m:
        m.setattr(Path, "resolve", _mock_existent_file)
        cp = config.GitlabConfigParser(gitlab_id="four")
    assert "four" == cp.gitlab_id
    assert "https://four.url" == cp.url
    assert cp.private_token is None
    assert "STUV" == cp.oauth_token
    assert 2 == cp.timeout
    assert cp.ssl_verify is True


@mock.patch("builtins.open")
def test_ssl_verify_as_str(m_open, monkeypatch):
    fd = io.StringIO(ssl_verify_str_config)
    fd.close = mock.Mock(return_value=None)
    m_open.return_value = fd

    with monkeypatch.context() as m:
        m.setattr(Path, "resolve", _mock_existent_file)
        cp = config.GitlabConfigParser()
    assert cp.ssl_verify == "/etc/ssl/certs/ca-certificates.crt"


@mock.patch("builtins.open")
@pytest.mark.skipif(sys.platform.startswith("win"), reason="Not supported on Windows")
def test_data_from_helper(m_open, monkeypatch, tmp_path):
    helper = tmp_path / "helper.sh"
    helper.write_text(
        dedent(
            """\
            #!/bin/sh
            echo "secret"
            """
        ),
        encoding="utf-8",
    )
    helper.chmod(0o755)

    fd = io.StringIO(
        dedent(
            f"""\
            [global]
            default = helper

            [helper]
            url = https://helper.url
            oauth_token = helper: {helper}
            """
        )
    )

    fd.close = mock.Mock(return_value=None)
    m_open.return_value = fd
    with monkeypatch.context() as m:
        m.setattr(Path, "resolve", _mock_existent_file)
        cp = config.GitlabConfigParser(gitlab_id="helper")
    assert "helper" == cp.gitlab_id
    assert "https://helper.url" == cp.url
    assert cp.private_token is None
    assert "secret" == cp.oauth_token


@mock.patch("builtins.open")
@pytest.mark.skipif(sys.platform.startswith("win"), reason="Not supported on Windows")
def test_from_helper_subprocess_error_raises_error(m_open, monkeypatch):
    # using false here to force a non-zero return code
    fd = io.StringIO(
        dedent(
            """\
            [global]
            default = helper

            [helper]
            url = https://helper.url
            oauth_token = helper: false
            """
        )
    )

    fd.close = mock.Mock(return_value=None)
    m_open.return_value = fd
    with monkeypatch.context() as m:
        m.setattr(Path, "resolve", _mock_existent_file)
        with pytest.raises(config.GitlabConfigHelperError) as e:
            config.GitlabConfigParser(gitlab_id="helper")

        assert "Failed to read oauth_token value from helper" in str(e.value)


@mock.patch("builtins.open")
@pytest.mark.parametrize(
    "config_string,expected_agent",
    [
        (valid_config, gitlab.const.USER_AGENT),
        (custom_user_agent_config, custom_user_agent),
    ],
)
def test_config_user_agent(m_open, monkeypatch, config_string, expected_agent):
    fd = io.StringIO(config_string)
    fd.close = mock.Mock(return_value=None)
    m_open.return_value = fd

    with monkeypatch.context() as m:
        m.setattr(Path, "resolve", _mock_existent_file)
        cp = config.GitlabConfigParser()
    assert cp.user_agent == expected_agent


@mock.patch("builtins.open")
@pytest.mark.parametrize(
    "config_string,expected",
    [
        pytest.param(valid_config, False, id="default_value"),
        pytest.param(
            global_retry_transient_errors(True), True, id="global_config_true"
        ),
        pytest.param(
            global_retry_transient_errors(False), False, id="global_config_false"
        ),
        pytest.param(
            global_and_gitlab_retry_transient_errors(False, True),
            True,
            id="gitlab_overrides_global_true",
        ),
        pytest.param(
            global_and_gitlab_retry_transient_errors(True, False),
            False,
            id="gitlab_overrides_global_false",
        ),
        pytest.param(
            global_and_gitlab_retry_transient_errors(True, True),
            True,
            id="gitlab_equals_global_true",
        ),
        pytest.param(
            global_and_gitlab_retry_transient_errors(False, False),
            False,
            id="gitlab_equals_global_false",
        ),
    ],
)
def test_config_retry_transient_errors_when_global_config_is_set(
    m_open, monkeypatch, config_string, expected
):
    fd = io.StringIO(config_string)
    fd.close = mock.Mock(return_value=None)
    m_open.return_value = fd

    with monkeypatch.context() as m:
        m.setattr(Path, "resolve", _mock_existent_file)
        cp = config.GitlabConfigParser()
    assert cp.retry_transient_errors == expected



"""
GitLab API: https://docs.gitlab.com/ee/api/resource_state_events.html
"""

import pytest
import responses

from gitlab.v4.objects import (
    ProjectIssueResourceStateEvent,
    ProjectMergeRequestResourceStateEvent,
)

issue_event_content = {"id": 1, "resource_type": "Issue"}
mr_event_content = {"id": 1, "resource_type": "MergeRequest"}


@pytest.fixture()
def resp_list_project_issue_state_events():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/issues/1/resource_state_events",
            json=[issue_event_content],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture()
def resp_get_project_issue_state_event():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/issues/1/resource_state_events/1",
            json=issue_event_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture()
def resp_list_merge_request_state_events():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/merge_requests/1/resource_state_events",
            json=[mr_event_content],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture()
def resp_get_merge_request_state_event():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/merge_requests/1/resource_state_events/1",
            json=mr_event_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_list_project_issue_state_events(
    project_issue, resp_list_project_issue_state_events
):
    state_events = project_issue.resourcestateevents.list()
    assert isinstance(state_events, list)

    state_event = state_events[0]
    assert isinstance(state_event, ProjectIssueResourceStateEvent)
    assert state_event.resource_type == "Issue"


def test_get_project_issue_state_event(
    project_issue, resp_get_project_issue_state_event
):
    state_event = project_issue.resourcestateevents.get(1)
    assert isinstance(state_event, ProjectIssueResourceStateEvent)
    assert state_event.resource_type == "Issue"


def test_list_merge_request_state_events(
    project_merge_request, resp_list_merge_request_state_events
):
    state_events = project_merge_request.resourcestateevents.list()
    assert isinstance(state_events, list)

    state_event = state_events[0]
    assert isinstance(state_event, ProjectMergeRequestResourceStateEvent)
    assert state_event.resource_type == "MergeRequest"


def test_get_merge_request_state_event(
    project_merge_request, resp_get_merge_request_state_event
):
    state_event = project_merge_request.resourcestateevents.get(1)
    assert isinstance(state_event, ProjectMergeRequestResourceStateEvent)
    assert state_event.resource_type == "MergeRequest"



"""
GitLab API: https://docs.gitlab.com/ce/api/repository_submodules.html
"""

import pytest
import responses


@pytest.fixture
def resp_update_submodule():
    content = {
        "id": "ed899a2f4b50b4370feeea94676502b42383c746",
        "short_id": "ed899a2f4b5",
        "title": "Message",
        "author_name": "Author",
        "author_email": "author@example.com",
        "committer_name": "Author",
        "committer_email": "author@example.com",
        "created_at": "2018-09-20T09:26:24.000-07:00",
        "message": "Message",
        "parent_ids": ["ae1d9fb46aa2b07ee9836d49862ec4e2c46fbbba"],
        "committed_date": "2018-09-20T09:26:24.000-07:00",
        "authored_date": "2018-09-20T09:26:24.000-07:00",
        "status": None,
    }

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.PUT,
            url="http://localhost/api/v4/projects/1/repository/submodules/foo%2Fbar",
            json=content,
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_update_submodule(project, resp_update_submodule):
    ret = project.update_submodule(
        submodule="foo/bar",
        branch="main",
        commit_sha="4c3674f66071e30b3311dac9b9ccc90502a72664",
        commit_message="Message",
    )
    assert isinstance(ret, dict)
    assert ret["message"] == "Message"
    assert ret["id"] == "ed899a2f4b50b4370feeea94676502b42383c746"
